#include <Arduino.h>
#include "OpenTherm.h"
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <Wire.h>         // Only needed for Arduino 1.6.5 and earlier
#include "SSD1306Wire.h"  // legacy include: `#include "SSD1306.h"`


SSD1306Wire display(0x3d, SDA, SCL);

#define VNV_DEF 4

// this part shall be changed ----------------------------------
IPAddress local_IP(0, 0, 0, 0);         //do to update
IPAddress gateway(0, 0, 0, 0);          //do to update
IPAddress subnet(0, 0, 0, 0);           //do to update
IPAddress mqtt_server(192, 168, x, x);  //do to update
String mqtt_client_id = "xxx";          // Unique client id
const char* ssid = "xxx";               //do to update
const char* password = "xxxxx";  //do to update
const int mqtt_Port = 1883;             //do to update
const char* mqtt_User = "xxx";          //do to update
const char* mqtt_Password = "xxx";      //do to update
//float maxVent = 2.49; //it means 249 m/h3 - max avilable flow in my Brink - not used/needed
const unsigned long readPeriod = 1000;           // 1000 = every second; set between 1000 - 5000
const unsigned long readPeriod_bypass = 120000;  // Set +15000 - OT disconnection needed for bypass work
const unsigned long readPeriodAll = 60000;

//-----------------------------------------------------------------------
const char* mqtt_topic_in = "brink/+/set";  //subscribe commands from Openhab

unsigned long startTime;
unsigned long currentTime;
unsigned long startTimeAll;

const int inPin = 4;   //ESP8266 D2
const int outPin = 5;  //ESP8266 D1
OpenTherm ot(inPin, outPin);
WiFiClient espClient;
PubSubClient mqttClient(espClient);

float tempSupplyIn, tempSupplyIn_old = -10000;
float tempExhaustOut, tempExhaustOut_old = -10000;
float tempSupplyOut, tempSupplyOut_old = -10000;
float tempExhaustIn, tempExhaustIn_old = -10000;
bool fault, fault_old = 1;
bool vmode, vmode_old = -10000;
int sbypass, sbypass_old = 1;
int pbypass, pbypass_old;
int filterDirty, filterDirty_old = 1;
int output, output_old = 0;
int input, input_old = 0;
int gvent, gvent_old = 0;
int pressin, pressin_old = -10000;
int pressout, pressout_old = -10000;  // uint8_t
int vstep1, vstep1_old = 0;
int vstep2, vstep2_old = 0;
int vstep3, vstep3_old = 0;
int vstep8, vstep8_old = 1;
float vstep9, vstep9_old = 0;
int byPassStatus, byPassStatus_old = 0;
int byPassAutoStatus, byPassAutoStatus_old = 0;
int tU4, tU4_old = 1;
float tU5, tU5_old = 1;
int cvol, cvol_old = 0;
int RPMin, RPMin_old;
int RPMout, RPMout_old;
int fcode, fcode_old = 0;
int msg, msg_old = 0;
int param1, param1_old;
long lRssi, lRssi_old = 0;
int cpod_old, cpod = 0;
float t, t_old = 0;
int h, h_old = 0;
//boolean retained = false;
const int publishDelay = 20;
long timeOn;
int datas;

void ICACHE_RAM_ATTR handleInterrupt() {
  ot.handleInterrupt();
}

void MqttReconnect() {
  while (!mqttClient.connected()) {
    Serial.print("Connecting to MQTT Broker: ");
    Serial.println(mqtt_server);
    delay(100);
    if (mqttClient.connect(mqtt_client_id.c_str(), mqtt_User, mqtt_Password)) {
      Serial.print("CONNECTED, topic: ");
      mqttClient.subscribe(mqtt_topic_in);
      Serial.println(mqtt_topic_in);
      Serial.println(mqttClient.state());
    } else {
      Serial.print("failed: ");
      Serial.print(mqttClient.state());
      Serial.println(" try again...");
      delay(5000);
    }
  }
  Serial.println(" ok...");
}

void MqttCallback(char* topic, byte* payload, unsigned int length) {
  payload[length] = 0;

  // Changing selected Brink Renovent parameters requested by OpenHab
  if (strcmp(topic, "brink/VentNomValue/set") == 0) ot.setVentilation(atoi((char*)payload));  // uint8_t
  if (strcmp(topic, "brink/U1/set") == 0) ot.setBrinkTSP(U1, atoi((char*)payload));
  if (strcmp(topic, "brink/U2/set") == 0) ot.setBrinkTSP(U2, atoi((char*)payload));
  if (strcmp(topic, "brink/U3/set") == 0) ot.setBrinkTSP(U3, atoi((char*)payload));
  if (strcmp(topic, "brink/I8/set") == 0) ot.setBrinkTSP(I8, atoi((char*)payload));  // bypass
  if (strcmp(topic, "brink/U4/set") == 0) ot.setBrinkTSP(U4, atoi((char*)payload) * 2);

  if (strcmp(topic, "brink/U5/set") == 0) {
    ot.setBrinkTSP(U5, atoi((char*)payload) * 2);
    refreshAll();  // change of U5 triggers refresh of other paramteres
  }

  if (strcmp(topic, "brink/REBOOT/set") == 0) {
    if (*payload == '1') {
      Serial.println("-=REBOOT=-");
      ESP.restart();
    }
  }

  if (strcmp(topic, "brink/SCAN/set") == 0) {
    String args = String((char*)payload);
    Serial.println("Scanning...");
    Serial.println((char*)payload);
    int start = 0;
    int stop = 72;
    bool brink = true;

    int i1 = args.indexOf(",");
    int i2 = args.indexOf(",", i1 + 1);


    if (i1 > 0) {
      if (args.substring(0, i1) == "o")
        brink = false;
      start = (args.substring(i1 + 1, i2)).toInt();
      if (i2 > 0) {
        stop = (args.substring(i2 + 1)).toInt();
      }
    }

    Serial.println("->" + String(i1) + "," + String(i2) + "=>" + String(start) + "," + String(stop));

    if (args.length() > 0) {
      Serial.println("-=SCAN=-");
      String s = "";
      String si = "";

      for (int i = start; i <= stop; i++) {
        si = scanner_TSP(i, brink);
        s += si;
        // scan(i);
        delay(20);
        if ((i > 0 && i % 5 == 0) || i == stop) {
          mqttClient.publish("brink/TestScan/get", s.c_str());
          Serial.println(";");
          s = "";
        }
      }
    }
  }
}

void setup() {
  delay(2500);
  Serial.begin(115200);
  Serial.println("Start");

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  if (WiFi.config(local_IP, gateway, subnet)) {
    Serial.println("Static IP Configured");
  } else {
    Serial.println("Static IP Configuration FAILED");
  }

  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(500);
  }
  Serial.println("");

  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
  Serial.print("MAC Address: ");
  Serial.println(WiFi.macAddress());

  mqtt_client_id = "brink_" + WiFi.macAddress();
  Serial.println(mqtt_client_id);

  mqttClient.setServer(mqtt_server, mqtt_Port);
  mqttClient.setKeepAlive(30);
  mqttClient.setBufferSize(1024);
  mqttClient.setCallback(MqttCallback);
  Serial.println(mqttClient.state());
  MqttReconnect();

  startTime = millis();
  ot.begin(handleInterrupt);

  timeOn = ((millis() / 1000) / 60);
  mqttClient.publish("brink/TimeON/get", String(timeOn).c_str());

  Serial.print("Number of available VentTSP ID's: ");
  Serial.println((int)ot.getTSPNumber());  // number of available VentTSP ID's

}

void updateTimeOn() {
  timeOn = ((millis() / 1000) / 60);
  mqttClient.publish("brink/TimeON/get", String(timeOn).c_str());
}

void loop() {
  double loopTime = millis();
  if (!mqttClient.connected()) {
    MqttReconnect();
    mqttClient.publish("brink/OpenThermStatus/get", "WORK");
  }
  mqttClient.loop();
  //ReadBrinkParameters();

  currentTime = millis();

  if (currentTime - startTime >= readPeriod) {
    ReadBrinkParameters();
    Monitor();
    lRssi = WiFi.RSSI();             //long
    if (abs(lRssi - lRssi_old) > 2)  //reduce data publication due frequent slight changes of signal
    {
      mqttClient.publish("brink/Wifi/get", String(lRssi).c_str());
      Serial.println(lRssi);
      lRssi_old = lRssi;
    }
    mqttClient.publish("brink/IPAddress/get", WiFi.localIP().toString().c_str());
    mqttClient.publish("brink/UniqueID/get", mqtt_client_id.c_str());
    startTime = currentTime;
  }

  loopTime = (millis() - loopTime) / 1000;
  if (loopTime > 1) {
    Serial.print("Loop took: ");
    Serial.print(loopTime);
    Serial.println(" seconds");
  }

  if (currentTime - startTimeAll >= readPeriodAll) {
    startTimeAll = currentTime;
    updateTimeOn();
    refreshAll();
  }
}

void publishValueF(String topic, float input) {
  if (input > -1000)
    mqttClient.publish(topic.c_str(), String(input).c_str());
}

void publishValue(String topic, long input) {
  if (input > -1000)
    mqttClient.publish(topic.c_str(), String(input).c_str());
}

float readAndPublishF(String topic, float input, float old) {
  if (input != old) {
    publishValueF(topic, input);
  }
  delay(publishDelay);
  return input;
}

int readAndPublish(String topic, int input, int old) {
  if (input != old) {
    publishValue(topic, input);
  }
  delay(publishDelay);
  return input;
}

//Reading selected Brink Renovent HR parameters, Mqtt publication only if a value has changed
void ReadBrinkParameters() {
  delay(publishDelay);

  tempSupplyIn_old = tempSupplyIn = readAndPublishF("brink/TempSupplyIn/get", ot.getVentSupplyInTemperature(), tempSupplyIn_old);
  tempSupplyOut_old = tempSupplyOut = readAndPublishF("brink/TempSupplyOut/get", ot.getVentSupplyOutTemperature(), tempSupplyOut_old);
  tempExhaustIn_old = tempExhaustIn = readAndPublishF("brink/TempExhIn/get", ot.getVentExhaustInTemperature(), tempExhaustIn_old);
  tempExhaustOut_old = tempExhaustOut = readAndPublishF("brink/TempExhOut/get", ot.getVentExhaustOutTemperature(), tempExhaustOut_old);

  filterDirty_old = filterDirty = readAndPublish("brink/FilterDirty/get", ot.getDiagnosticIndication(), filterDirty_old);

  gvent_old = gvent = readAndPublish("brink/VentilationNominalValue/get", ot.getVentilation() * VNV_DEF, gvent_old);
  int cvolx = ot.getBrinkTSP(CurrentVol);
  if (gvent > 255 && cvolx < 100) 
    cvolx = cvolx + 255;
  cvol_old = cvol = readAndPublish("brink/CurrentVolume/get", cvolx, cvol_old);

  int coutv = ot.getBrinkTSP(CurrentOutputVol);
  if (gvent > 255 && coutv < 100) 
    coutv = coutv + 255;
  output_old = output = readAndPublish("brink/CurrentOutputVol/get", coutv, output_old);

  int cintv = ot.getBrinkTSP(CurrentInputVol2);
  if (gvent > 255 && cintv < 100) 
    cintv = cintv + 255;
  input_old = input = readAndPublish("brink/CurrentInputVol/get", cintv, input_old);
  
  pressin_old = pressin = readAndPublish("brink/CPID/get", ot.getBrinkTSP(CPID), pressin_old);
  pressout_old = pressout = readAndPublish("brink/CPOD/get", ot.getBrinkTSP(CPOD), pressout_old);
  
  byPassStatus_old = byPassStatus = readAndPublish("brink/BypassStatus/get", ot.getBrinkTSP(BypassStatus), byPassStatus_old);
  vstep1_old = vstep1 = readAndPublish("brink/U1/get", ot.getBrinkTSP(U1), vstep1_old);
  vstep2_old = vstep2 = readAndPublish("brink/U2/get", ot.getBrinkTSP(U2), vstep2_old);
  vstep3_old = vstep3 = readAndPublish("brink/U3/get", ot.getBrinkTSP(U3), vstep3_old);
  tU4_old = tU4 = readAndPublish("brink/U4/get", ot.getBrinkTSP(U4) /2, tU4_old);
  tU5_old = tU5 = readAndPublish("brink/U5/get", ot.getBrinkTSP(U5) /2, tU5_old);
  param1_old = param1 = readAndPublish("brink/I5/get", ot.getBrinkTSP(I5), param1_old);

  fault_old = fault = readAndPublish("brink/FaultIndication/get", ot.getFaultIndication(), fault_old);
  vmode_old = vmode = readAndPublish("brink/VentilationMode/get", ot.getVentilationMode(), vmode_old);
  byPassAutoStatus_old = byPassAutoStatus = readAndPublish("brink/BypassAutomaticStatus/get", ot.getBypassAutomaticStatus(), byPassAutoStatus_old);

}


// refresh slow changing parameters on request by every change of U5
void refreshAll() {
  publishValueF("brink/TempSupplyIn/get", tempSupplyIn);
  publishValueF("brink/TempSupplyOut/get", tempSupplyOut);
  publishValueF("brink/TempExhIn/get", tempExhaustIn);
  publishValueF("brink/TempExhOut/get", tempExhaustOut);
  publishValue("brink/FaultIndication/get", fault);
  publishValue("brink/VentilationMode/get", vmode);
  publishValue("brink/BypassStatus/get", sbypass);
  publishValue("brink/BypassAutomaticStatus/get", byPassAutoStatus);
  publishValue("brink/FilterDirty/get", filterDirty);
  publishValue("brink/VentilationNominalValue/get", gvent);
  publishValue("brink/CurrentVolume/get", cvol);
  publishValue("brink/FaultCode/get", fcode);
  publishValue("brink/OperationMsg/get", msg);
  publishValue("brink/I1/get", param1 - 100);
  publishValue("brink/U1/get", vstep1);
  publishValue("brink/U2/get", vstep2);
  publishValue("brink/U3/get", vstep3);
  publishValue("brink/U4/get", tU4);
  publishValue("brink/U5/get", tU5);
  publishValue("brink/Wifi/get", lRssi);
  mqttClient.publish("brink/OpenThermStatus/get", "WORK");
}

/*

  // mqttClient.publish("brink/MinVol/get", String(ot.getBrinkTSP(MaxVol)).c_str());
*/

String scanner_TSP(int id, bool brinkTSP) {
  Serial.print("." + String(id));
  String s = "";
  if (brinkTSP) {
    int b8 = ot.getBrinkTSP(static_cast<BrinkTSPindex>(id));
    int b16 = ot.getBrink2TSP(static_cast<BrinkTSPindex>(id));
    float f = ot.getFloat(b16);
    s = "[brink" + String(id) + ":" + String(b8) + "," + String(b16) + "," + String(f) + "]";
  } else {
    unsigned long response = ot.sendRequest(ot.buildRequest(OpenThermRequestType::READ, static_cast<OpenThermMessageID>(id), 0));
    int b8 = ot.getU8(response);
    int b16 = ot.getUInt(response);
    float f = ot.getFloat(b16);
    s = "[ot" + String(id) + ":" + String(b8) + "," + String(b16) + "," + String(f) + "]";
  }

  return s;
}

void scan(int id) {
  unsigned long msg = ot.buildRequest(OpenThermRequestType::READ, (OpenThermMessageID)id, 0);
  unsigned long response = ot.sendRequest(msg);

  if (ot.isValidResponse(response)) {
    unsigned int data = ot.getUInt(response);
    Serial.print("ID: ");
    Serial.print(id);
    Serial.print(" ==> DATA: ");
    Serial.println(data);
  } else {
    //Serial.println("INVALID");
  }
}

void Monitor() {
  Serial.print(String(cvol).c_str());
  Serial.print(", ");
  Serial.print(String(gvent * 4.5).c_str());
  Serial.print(", ");
  Serial.print(String(output).c_str());
  Serial.print(", ");
  Serial.print(String(input).c_str());
  Serial.print(", ");
  Serial.print(String(tempSupplyIn).c_str());
  Serial.print(", ");
  Serial.print(String(tempExhaustOut).c_str());
  Serial.print(", ");
  Serial.print(String(filterDirty).c_str());
  Serial.print(", ");
  Serial.print(String(vstep1).c_str());
  Serial.print(", ");
  Serial.print(String(vstep2).c_str());
  Serial.print(", ");
  Serial.print(String(vstep3).c_str());
  Serial.print(", ");
  Serial.print(String(tU4 / 2).c_str());
  Serial.print(", ");
  Serial.print(String(tU5 / 2).c_str());
  Serial.print(", ");
  Serial.print(String(pbypass).c_str());
  Serial.print(", ");
  Serial.print(String(sbypass).c_str());
  Serial.print(", ");
  Serial.println(String(byPassStatus).c_str());
  Serial.print("==== ");
  Serial.println(String(cpod).c_str());
}
